"""useful model for JSON Table Schema."""
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import warnings
from . import types
from . import helpers
from . import exceptions
from .validate import validate


class SchemaModel(object):

    """Model for a JSON Table Schema.

    Providers handy helpers for ingesting, validating and outputting
    JSON Table Schemas: http://dataprotocols.org/json-table-schema/

    Args:
        * schema_source (string or dict): A filepath, url or dictionary
        that represents a schema
        * case_insensitive_headers (bool): if True, headers should be
        considered case insensitive, and `SchemaModel` forces all
        headers to lowercase when they are represented via a model
        instance. This setting **does not** mutate the actual strings
        that come from the the input schema_source, so out put methods
        such as as_python and as_json are **not** subject to this flag.

    """

    NULL_VALUES = [None] + helpers.NULL_VALUES
    TRUE_VALUES = [True] + helpers.TRUE_VALUES
    FALSE_VALUES = [False] + helpers.FALSE_VALUES

    DEFAULTS = {
        'format': 'default',
        'type': 'string'
    }

    def __init__(self, schema_source, case_insensitive_headers=False):

        # DEPRECATED [v0.7-v1)
        message = 'Class models.SchemaModel is deprecated [v0.7-v1)'
        warnings.warn(message, UserWarning)

        self.schema_source = schema_source
        self.case_insensitive_headers = case_insensitive_headers
        _as_python = self._to_python()

        if _as_python is None:
            raise exceptions.InvalidJSONError

        try:
            validate(_as_python)
        except exceptions.SchemaValidationError:
            raise exceptions.InvalidSchemaError

        self.as_python = self._expand(_as_python)
        self.as_json = json.dumps(self.as_python)

    @property
    def headers(self):
        _raw = [f['name'] for f in self.as_python.get('fields')]
        if self.case_insensitive_headers:
            return [name.lower() for name in _raw]
        return _raw

    @property
    def required_headers(self):
        _raw = [f['name'] for f in self.as_python.get('fields')
                if f.get('constraints', {}).get('required')]
        if self.case_insensitive_headers:
            return [name.lower() for name in _raw]
        return _raw

    @property
    def primaryKey(self):
        return self.as_python.get('primaryKey')

    @property
    def foreignKeys(self):
        return self.as_python.get('foreignKeys')

    @property
    def fields(self):
        return [f for f in self.as_python.get('fields')]

    def cast(self, field_name, value, index=0):
        """Return the cast value or raise an exception if value can't be cast
        as the field_name's type."""

        _type = self.get_type(field_name, index=index)

        return _type.cast(value)

    def get_field(self, field_name, index=0):
        """Return the `field` object for `field_name`.

        `index` allows accessing a field name by position, as JTS allows
        duplicate field names.

        """

        try:
            return [f for f in self.fields if f['name'] == field_name][index]
        except IndexError:
            return None

    def has_field(self, field_name):
        """Return boolean if the field exists in the schema."""

        return bool(self.get_field(field_name))

    def get_type(self, field_name, index=0):
        """Return the `type` for `field_name`."""

        _field = self.get_field(field_name, index=index)
        _class = self._type_map()[_field['type']]

        return _class(_field)

    def get_fields_by_type(self, type_name):
        """Return all fields that match the given type."""

        return [f for f in self.fields if f['type'] == type_name]

    def get_constraints(self, field_name, index=0):
        """Return the `constraints` object for `field_name`."""

        return self.get_field(field_name, index=index).get('constraints')

    def _to_python(self):
        """Return schema as a Python data structure (dict)."""

        try:
            return helpers.load_json_source(self.schema_source)
        except Exception:
            return None

    def _expand(self, schema):
        """Expand the schema with additional default properties."""

        for field in schema.get('fields', {}):

            # ensure we have a default type if no type was declared
            if not field.get('type'):
                field['type'] = self.DEFAULTS['type']

            # ensure we have a default format if no format was declared
            if not field.get('format'):
                field['format'] = self.DEFAULTS['format']

        return schema

    @staticmethod
    def _type_map():
        """Map a JSON Table Schema type to a JTSKit type class."""

        return {
            'string': types.StringType,
            'number': types.NumberType,
            'integer': types.IntegerType,
            'boolean': types.BooleanType,
            'null': types.NullType,
            'array': types.ArrayType,
            'object': types.ObjectType,
            'date': types.DateType,
            'time': types.TimeType,
            'datetime': types.DateTimeType,
            'geopoint': types.GeoPointType,
            'geojson': types.GeoJSONType,
            'any': types.AnyType
        }

    def convert_row(self, *items, **kwargs):
        fail_fast = kwargs.pop('fail_fast', False)
        if len(self.headers) != len(items):
            raise exceptions.ConversionError(
                'The number of items to convert does not match the number of '
                'fields given in the schema\n'
                'headers : {0} - {1}\nitems : {2} - {3}'.format(
                    len(self.headers), self.headers, len(items), items
                )
            )
        errors = []
        for field_name, item in zip(self.headers, items):
            try:
                yield self.cast(field_name, item)
            except exceptions.InvalidCastError as e:
                if fail_fast:
                    raise
                else:
                    errors.append(e)
        if errors:
            raise exceptions.MultipleInvalid(errors=errors)

    def convert(self, rows, fail_fast=False):
        errors = []
        for row in rows:
            try:
                yield list(self.convert_row(*row, fail_fast=fail_fast))
            except exceptions.MultipleInvalid as e:
                errors.extend(e.errors)
            except exceptions.ConversionError as e:
                if fail_fast:
                    raise
                else:
                    errors.append(e)
        if errors:
            raise exceptions.MultipleInvalid(errors=errors)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import json
from copy import deepcopy
from .field import Field
from .validate import validate
from . import compat
from . import helpers
from . import exceptions


# Module API

class Schema(object):
    """JSON Table Schema schema representation.

    Raises:
        exceptions.InvalidJSONError
        exceptions.SchemaValidationError

    Args:
        descriptor (dict/str): schema descriptor/path/url

    """

    # Public

    def __init__(self, descriptor):

        # Set attributes
        self.__descriptor = deepcopy(helpers.load_json_source(descriptor))
        self.__fields = None

        # Validate
        validate(self.__descriptor)

    @property
    def descriptor(self):
        """dict: schema descriptor
        """
        return self.__descriptor

    def cast_row(self, row, no_fail_fast=False):
        """Cast row to schema types.

        Args:
            row (mixed[]): array of values
            no_fail_fast (bool): collect all error

        Raises:
            exceptions.InvalidCastError
            exceptions.MultipleInvalid (no_fail_fast=True)

        Returns:
            mixed[]: cast row

        """

        # Prepare
        errors = []

        # Check row length
        if len(row) != len(self.fields):
            message = 'Row length (%s) doesn\'t match fields count (%s)'
            message = message % (len(row), len(self.fields))
            exception = exceptions.InvalidCastError(message)
            if not no_fail_fast:
                raise exception
            errors.append(exception)

        # Cast
        result = []
        if not errors:
            for field, value in zip(self.fields, row):
                try:
                    result.append(field.cast_value(value))
                except exceptions.InvalidCastError as exception:
                    if not no_fail_fast:
                        raise
                    errors.append(exception)

        # Raise
        if errors:
            raise exceptions.MultipleInvalid(errors=errors)

        return result

    @property
    def fields(self):
        """Field[]: field instances
        """
        if self.__fields is None:
            self.__fields = [Field(fd) for fd in self.__descriptor['fields']]
        return self.__fields

    def get_field(self, name):
        """Return field by name.

        Args:
            name (str): field name

        Returns:
            None/Field: return field instance or
                None if field with name doesn't exist

        """
        for field in self.fields:
            if field.name == name:
                return field
        return None

    def has_field(self, name):
        """Check if field exists.

        Args:
            name (str): field name

        Returns:
            bool: existence of field

        """
        return self.get_field(name) is not None

    @property
    def headers(self):
        """str[]: field names (headers)
        """
        return [field.name for field in self.fields]

    @property
    def primary_key(self):
        """str[]: primary key
        """
        primary_key = self.__descriptor.get('primaryKey', [])
        if isinstance(primary_key, compat.str):
            primary_key = [primary_key]
        return primary_key

    @property
    def foreign_keys(self):
        """dict[]: foreign keys
        """
        return self.__descriptor.get('foreignKeys', [])

    def save(self, target):
        """Save schema descriptor.

        Args:
            target (str): file path

        """
        mode = 'w'
        encoding = 'utf-8'
        if compat.is_py2:
            mode = 'wb'
            encoding = None
        helpers.ensure_dir(target)
        with io.open(target, mode=mode, encoding=encoding) as file:
            json.dump(self.__descriptor, file, indent=4)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .schema import Schema
from .field import Field
from .table import Table
from .storage import Storage
from .validate import validate
from .infer import infer
from . import exceptions
from . import types

# Deprecated
from .validate import validator
from .pushpull import push_resource, pull_resource
from . import model
from . import storage



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


# Module API

def check_minLength(name, value, min_length):
    """Length constraint.

    Supported types: sting, array, object.

    Args:
        name (str): field name
        value (str/list/dict): field value
        min_length (str/list/dict): min length to check

    Raises:
        TypeError: for non supported type
        ConstraintError: if check is failed

    """
    if len(value) < min_length:
        raise exceptions.ConstraintError(
            "The field '{0}' must have a minimum length of {1}"
            .format(name, min_length))
    return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time
import datetime
from future.utils import raise_with_traceback
from dateutil.parser import parse as date_parse
from .. import exceptions
from .. import helpers
from . import base


# Module API

class TimeType(base.JTSType):

    # Public

    name = 'time'
    null_values = helpers.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minimum',
        'maximum',
    ]
    # ---
    python_type = datetime.time
    ISO8601 = '%H:%M:%S'
    raw_formats = ['HH/MM/SS']
    py_formats = ['%H:%M:%S']
    format_map = dict(zip(raw_formats, py_formats))

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            struct_time = time.strptime(value, self.ISO8601)
            return datetime.time(
                struct_time.tm_hour, struct_time.tm_min, struct_time.tm_sec)
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidTimeType(e))

    def cast_any(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return date_parse(value).time()
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidTimeType(e))

    def cast_fmt(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return datetime.datetime.strptime(value, fmt).time()
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidTimeType(e))



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import deepcopy
from . import types


# Module API

class Field(object):
    """JSON Table Schema field representation.
    """

    # Public

    def __init__(self, descriptor):
        self.__descriptor = deepcopy(descriptor)
        # Probably it's just a temporal solution
        self.__type = _TYPES[self.type](descriptor)

    @property
    def descriptor(self):
        return self.__descriptor

    @property
    def name(self):
        return self.__descriptor['name']

    @property
    def type(self):
        return self.__descriptor.get('type', 'string')

    @property
    def format(self):
        return self.__descriptor.get('format', 'default')

    @property
    def constraints(self):
        return self.__descriptor.get('constraints', {})

    def cast_value(self, value):
        return self.__type.cast(value)


# Internal

_TYPES = {
    'string': types.StringType,
    'number': types.NumberType,
    'integer': types.IntegerType,
    'boolean': types.BooleanType,
    'null': types.NullType,
    'array': types.ArrayType,
    'object': types.ObjectType,
    'date': types.DateType,
    'time': types.TimeType,
    'datetime': types.DateTimeType,
    'geopoint': types.GeoPointType,
    'geojson': types.GeoJSONType,
    'any': types.AnyType,
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import decimal
from future.utils import raise_with_traceback
from .. import exceptions
from .. import helpers
from .. import compat
from . import base


# Module API

class GeoPointType(base.JTSType):

    # Public

    name = 'geopoint'
    null_values = helpers.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
    ]
    # ---
    python_types = (compat.str, list, dict)

    def cast_default(self, value, fmt=None):
        try:
            if isinstance(value, self.python_types):
                points = value.split(',')
                if len(points) == 2:
                    try:
                        geopoints = [decimal.Decimal(points[0].strip()),
                                     decimal.Decimal(points[1].strip())]
                        self.__check_latitude_longtiude_range(geopoints)
                        return geopoints
                    except decimal.DecimalException as e:
                        raise_with_traceback(exceptions.InvalidGeoPointType(e))
                else:
                    raise exceptions.InvalidGeoPointType(
                        '{0}: point is not of length 2'.format(value))
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidGeoPointType(e))

    def cast_array(self, value, fmt=None):
        try:
            json_value = json.loads(value)
            if isinstance(json_value, list) and len(json_value) == 2:
                try:
                    longitude = json_value[0].strip()
                    latitude = json_value[1].strip()
                except AttributeError:
                    longitude = json_value[0]
                    latitude = json_value[1]

                try:
                    geopoints = [decimal.Decimal(longitude),
                                 decimal.Decimal(latitude)]
                    self.__check_latitude_longtiude_range(geopoints)
                    return geopoints
                except decimal.DecimalException as e:
                    raise_with_traceback(exceptions.InvalidGeoPointType(e))
            else:
                raise exceptions.InvalidGeoPointType(
                    '{0}: point is not of length 2'.format(value))
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidGeoPointType(e))

    def cast_object(self, value, fmt=None):
        try:
            json_value = json.loads(value)

            try:
                longitude = json_value['longitude'].strip()
                latitude = json_value['latitude'].strip()
            except AttributeError:
                longitude = json_value['longitude']
                latitude = json_value['latitude']
            except KeyError as e:
                raise_with_traceback(exceptions.InvalidGeoPointType(e))

            try:
                geopoints = [decimal.Decimal(longitude),
                             decimal.Decimal(latitude)]
                self.__check_latitude_longtiude_range(geopoints)
                return geopoints
            except decimal.DecimalException as e:
                raise_with_traceback(exceptions.InvalidGeoPointType(e))
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidGeoPointType(e))

    # Private

    def __check_latitude_longtiude_range(self, geopoint):
        longitude = geopoint[0]
        latitude = geopoint[1]
        if longitude >= 180 or longitude <= -180:
            raise exceptions.InvalidGeoPointType(
                'longtitude should be between -180 and 180, '
                'found: {0}'.format(longitude)
            )
        elif latitude >= 90 or latitude <= -90:
            raise exceptions.InvalidGeoPointType(
                'latitude should be between -90 and 90, '
                'found: {0}'.format(latitude))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import json
import warnings
import jsonschema
from jsonschema.validators import validator_for
from . import compat, exceptions


# Module API

def validate(descriptor, no_fail_fast=False):
    """Validate JSON Table Schema schema descriptor.

    Args:
        descriptor (dict): schema descriptor to validate
        no_fail_fast (bool): collect all errors

    Raises:
        exceptions.SchemaValidationError
        exceptions.MultipleInvalid (no_fail_fast=True)

    Returns:
        bool: True

    """

    # Fail fast
    if not no_fail_fast:
        jsonschema.validate(
            descriptor, _json_table_schema,
            cls=_JSONTableSchemaValidator)

    # Multiple errors
    else:
        validator = _JSONTableSchemaValidator(_json_table_schema)
        errors = list(validator.iter_errors(descriptor))
        if errors:
            raise exceptions.MultipleInvalid(errors=errors)

    return True


# Deprecated
class Validator(object):
    @staticmethod
    def iter_errors(schema):
        # DEPRECATED [v0.7-v1)
        message = 'validator is deprecated [v0.7-v1)'
        warnings.warn(message, UserWarning)
        validator = _JSONTableSchemaValidator(_json_table_schema)
        return validator.iter_errors(schema)
validator = Validator()


# Internal

# Get schema and validator
def _load_schema_and_validator():
    basepath = os.path.dirname(__file__)
    filepath = os.path.join(basepath, 'schemas/json-table-schema.json')
    with open(filepath) as file:
        json_table_schema = json.load(file)
    BaseValidator = validator_for(json_table_schema)
    return json_table_schema, BaseValidator
_json_table_schema, _BaseValidator = _load_schema_and_validator()


class _JSONTableSchemaValidator(_BaseValidator):
    @classmethod
    def check_schema(cls, schema):
        # When checking against the metaschema, we do not want to run the
        # additional checking added in iter_errors
        parent_cls = cls.__bases__[0]
        for error in parent_cls(cls.META_SCHEMA).iter_errors(schema):
            raise jsonschema.exceptions.SchemaError.create_from(error)

    def iter_errors(self, instance, _schema=None):

        for e in super(_JSONTableSchemaValidator, self).iter_errors(
                instance, _schema):
            yield exceptions.SchemaValidationError(
                e.message, e.validator, e.path, e.cause, e.context,
                e.validator_value, e.instance, e.schema, e.schema_path,
                e.parent)

        try:
            field_names = [f['name'] for f in instance['fields']]
        except (TypeError, KeyError):
            field_names = []
        # the hash MAY contain a key `primaryKey`
        if isinstance(instance, dict) and instance.get('primaryKey'):
            # ensure that the primary key matches field names

            if isinstance(instance['primaryKey'], compat.str):
                if not instance['primaryKey'] in field_names:
                    yield exceptions.SchemaValidationError(
                        'A JSON Table Schema primaryKey value must be found in'
                        ' the schema field names')
            elif isinstance(instance['primaryKey'], list):
                for k in instance['primaryKey']:
                    if k not in field_names:
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema primaryKey value must be '
                            'found in the schema field names')

        # the hash may contain a key `foreignKeys`
        if isinstance(instance, dict) and instance.get('foreignKeys'):
            for fk in instance['foreignKeys']:
                # ensure that `foreignKey.fields` match field names
                if isinstance(fk.get('fields'), compat.str):
                    if fk.get('fields') not in field_names:
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.fields value must '
                            'correspond with field names.')
                elif isinstance(fk.get('fields', []), list):
                    for field in fk.get('fields'):
                        if field not in field_names:
                            yield exceptions.SchemaValidationError(
                                'A JSON Table Schema foreignKey.fields value '
                                'must correspond with field names.')

                # ensure that `foreignKey.reference.fields`
                # matches outer `fields`
                if isinstance(fk.get('fields'), compat.str):
                    if not isinstance(fk['reference']['fields'], compat.str):
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.reference.fields '
                            'must match field names.')
                else:
                    if isinstance(fk['reference']['fields'], compat.str):
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.fields cannot '
                            'be a string when foreignKey.reference.fields.'
                            'is a string')
                    if not (len(fk.get('fields')) ==
                            len(fk['reference']['fields'])):
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.fields must '
                            'contain the same number entries as '
                            'foreignKey.reference.fields.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import uuid
import base64
import decimal
import binascii
from rfc3986 import is_valid_uri
from future.utils import raise_with_traceback
from .. import exceptions
from .. import helpers
from .. import compat
from . import base


# Module API

class StringType(base.JTSType):

    # Public

    name = 'string'
    null_values = [value for value in helpers.NULL_VALUES if value != '']
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minLength',
        'maxLength',
    ]
    # ---
    python_type = compat.str
    email_pattern = re.compile(r'[^@]+@[^@]+\.[^@]+')

    def cast_default(self, value, fmt=None):

        if not isinstance(value, self.python_type):
            raise exceptions.InvalidStringType(
                '{0} is not of type {1}'.format(value, self.python_type))

        try:
            return self.python_type(value)
        except (ValueError, TypeError, decimal.InvalidOperation) as e:
            raise_with_traceback(exceptions.InvalidCastError(e))

        raise exceptions.InvalidCastError('Could not cast value')

    def cast_email(self, value, fmt=None):

        if not isinstance(value, self.python_type):
            raise exceptions.InvalidStringType(
                '{0} is not of type {1}'.format(value, self.python_type))

        if not re.match(self.email_pattern, value):
            raise exceptions.InvalidEmail(
                '{0} is not a valid email'.format(value))

        return value

    def cast_uri(self, value, fmt=None):

        if not isinstance(value, self.python_type):
            raise exceptions.InvalidStringType(
                '{0} is not of type {1}'.format(value, self.python_type))

        if is_valid_uri(value, require_scheme=True):
            return value
        else:
            raise exceptions.InvalidURI(
                '{0} is not a valid uri'.format(value))

    def cast_binary(self, value, fmt=None):

        if not isinstance(value, self.python_type):
            raise exceptions.InvalidStringType(
                '{0} is not of type {1}'.format(value, self.python_type))

        try:
            base64.b64decode(value)
        except binascii.Error as e:
            raise_with_traceback(exceptions.InvalidBinary(e))
        return value

    def cast_uuid(self, value, fmt=None):

        if not isinstance(value, self.python_type):
            raise exceptions.InvalidStringType(
                '{0} is not of type {1}'.format(value, self.python_type))

        try:
            uuid.UUID(value, version=4)
            return value
        except ValueError as e:
            raise_with_traceback(exceptions.InvalidUUID(e))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import decimal
from future.utils import raise_with_traceback
from .. import exceptions
from .. import helpers
from . import base


# Module API

class IntegerType(base.JTSType):

    # Public

    name = 'integer'
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minimum',
        'maximum',
    ]
    null_values = helpers.NULL_VALUES
    # ---
    python_type = int

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return self.python_type(value)
        except (ValueError, TypeError, decimal.InvalidOperation) as e:
            raise_with_traceback(exceptions.InvalidCastError(e))

        raise exceptions.InvalidCastError('Could not cast value')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import operator
from . import types


# Module API

def infer(headers, values, row_limit=None, explicit=False, primary_key=None):
    """Return a schema from the passed headers and values.

    Args:
        headers: a list of header names
        values: a reader over data, yielding each row as a list of values
        explicit: be explicit.
        primary_key: pass in a primary key or iterable of keys.

    Returns:
        A JSON Table Schema as a Python dict.

    """

    guesser = _TypeGuesser()
    resolver = _TypeResolver()
    schema = {'fields': []}
    type_matches = {}

    if primary_key:
        schema['primaryKey'] = primary_key

    for header in headers:
        descriptor = {
            'name': header,
            'title': '',
            'description': '',
        }

        constraints = {}
        if explicit:
            constraints.update({'required': True})

        if header == primary_key:
            constraints.update({'unique': True})

        if constraints:
            descriptor['constraints'] = constraints

        schema['fields'].append(descriptor)

    for index, row in enumerate(values):

        if row_limit and (index > row_limit):
            break

        else:

            # Normalize rows with invalid dimensions for sanity
            row_length = len(row)
            headers_length = len(headers)

            if row_length > headers_length:
                row = row[:len(headers)]

            if row_length < headers_length:
                diff = headers_length - row_length
                fill = [''] * diff
                row = row + fill

            # build a column-wise lookup of type matches
            for index, value in enumerate(row):
                rv = guesser.cast(value)

                if type_matches.get(index):
                    type_matches[index].append(rv)
                else:
                    type_matches[index] = [rv]

    # choose a type/format for each column based on the matches
    for index, results in type_matches.items():
        rv = resolver.get(results)
        schema['fields'][index].update(**rv)

    return schema


# Internal

class _TypeGuesser(object):
    """Guess the type for a value.

    Returns:
        A tuple  of ('type', 'format')

    """

    def __init__(self, type_options=None):
        self._types = _get_available_types()
        self.type_options = type_options or {}

    def cast(self, value):
        for _type in reversed(self._types):
            result = _type(self.type_options.get(_type.name, {})).test(value)
            if result:
                rv = (_type.name, 'default')
                break

        return rv


class _TypeResolver(object):
    """Get the best matching type/format from a list of possible ones.
    """

    def __init__(self):
        self._types = _get_available_types()

    def get(self, results):

        variants = set(results)

        # only one candidate... that's easy.
        if len(variants) == 1:
            rv = {
                'type': results[0][0],
                'format': results[0][1],
            }

        else:
            counts = {}
            for result in results:
                if counts.get(result):
                    counts[result] += 1
                else:
                    counts[result] = 1

            # tuple representation of `counts` dict sorted by values
            sorted_counts = sorted(counts.items(), key=operator.itemgetter(1),
                                   reverse=True)
            rv = {
                'type': sorted_counts[0][0][0],
                'format': sorted_counts[0][0][1]
            }

        return rv


def _get_available_types():
    """Return available types.
    """
    return [
        types.AnyType,
        types.StringType,
        types.BooleanType,
        types.NumberType,
        types.IntegerType,
        types.NullType,
        types.DateType,
        types.TimeType,
        types.DateTimeType,
        types.ArrayType,
        types.ObjectType,
        types.GeoPointType,
        types.GeoJSONType,
    ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
from future.utils import raise_with_traceback
from dateutil.parser import parse as date_parse
from .. import exceptions
from .. import helpers
from . import base


# Module API

class DateType(base.JTSType):

    # Public

    name = 'date'
    null_values = helpers.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minimum',
        'maximum',
    ]
    # ---
    python_type = datetime.date
    formats = ('default', 'any', 'fmt')
    ISO8601 = '%Y-%m-%d'
    raw_formats = ['DD/MM/YYYY', 'DD/MM/YY', 'YYYY/MM/DD']
    py_formats = ['%d/%m/%Y', '%d/%m/%y', '%Y/%m/%d']
    format_map = dict(zip(raw_formats, py_formats))

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return datetime.datetime.strptime(value, self.ISO8601).date()
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidDateType(e))

    def cast_any(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return date_parse(value).date()
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidDateType(e))

    def cast_fmt(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return datetime.datetime.strptime(value, fmt).date()
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidDateType(e))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import io
import json
import click
import jsontableschema


DIR = os.path.abspath(os.path.dirname(__file__))
VERSION_FILE = os.path.join(os.path.dirname(__file__), 'VERSION')
VERSION = io.open(VERSION_FILE, encoding='utf-8').read().strip()


@click.group()
def main():
    """The entry point into the CLI."""


@main.command()
def info():
    """Return info on this version of JSON Table Schema"""
    click.echo(json.dumps({'version': VERSION}, ensure_ascii=False, indent=4))


@main.command()
@click.argument('data')
@click.option('--row_limit', default=0, type=int)
@click.option('--encoding', default='utf-8')
@click.option('--to_file')
def infer(data, row_limit, encoding, to_file):

    """Infer a schema from data.

    * data must be a local filepath
    * data must be CSV
    * the file encoding is assumed to be UTF-8 unless an encoding is passed
      with --encoding
    * the first line of data must be headers
    * these constraints are just for the CLI

    """

    if not row_limit:
        row_limit = None

    with io.open(data, mode='r+t', encoding=encoding) as stream:
        try:
            headers = stream.readline().rstrip('\n').split(',')
            values = jsontableschema.compat.csv_reader(stream)
        except UnicodeDecodeError:
            response = "Could not decode the data file as {0}. " \
                "Please specify an encoding to use with the " \
                "--encoding argument.".format(encoding)
        else:
            response = jsontableschema.infer(headers, values,
                                             row_limit=row_limit)

        if to_file:
            with io.open(to_file, mode='w+t', encoding='utf-8') as dest:
                dest.write(json.dumps(response, ensure_ascii=False, indent=2))

    click.echo(response)


@main.command()
@click.argument('schema')
def validate(schema):

    """Validate that a supposed schema is in fact a JSON Table Schema."""

    errors = [e.message for e in jsontableschema.validator.iter_errors(schema)]
    if not errors:
        click.echo(False)
    else:
        click.echo(True)
        click.echo(errors)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


# Module API

def check_minimum(name, value, minimum):
    """Minimum and maximum constraint.

    Supported types: integer, number, datetime, date, time.

    Args:
        name (str): field name
        value (int/float/Decimal/datetime/date/time): field value
        minimum (int/float/Decimal/datetime/date/time): minimum to check

    Raises:
        TypeError: for non supported type
        ConstraintError: if check is failed

    """
    if value < minimum:
        raise exceptions.ConstraintError(
            "The field '{0}' must not be less than {1}"
            .format(name, minimum))
    return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import helpers
from . import base


# Module API

class AnyType(base.JTSType):

    # Public

    name = 'any'
    null_values = helpers.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
    ]

    def cast_default(self, value, fmt=None):
        return value



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import jsonschema.exceptions


# Base

class JsonTableSchemaException(Exception):
    pass


class MultipleInvalid(JsonTableSchemaException):
    def __init__(self, msg='Multiple errors found', errors=None):
        self.msg = msg
        if errors:
            self.errors = errors
        else:
            self.errors = []
        super(MultipleInvalid, self).__init__(msg)


# Load

class InvalidJSONError(Exception):
    def __init__(self, msg=None):
        self.msg = msg or 'The obj cannot be parsed as JSON.'


# Validate

class InvalidSchemaError(Exception):
    def __init__(self, msg=None):
        self.msg = msg or 'The obj is not a valid JSON Table Schema.'


class SchemaValidationError(
        JsonTableSchemaException,
        jsonschema.exceptions.ValidationError):
    pass


# Cast

class InvalidCastError(JsonTableSchemaException):
    pass


class InvalidStringType(InvalidCastError):
    pass


class IncorrectTypeError(InvalidStringType):
    pass


class InvalidEmail(InvalidStringType):
    pass


class InvalidURI(InvalidStringType):
    pass


class InvalidBinary(InvalidStringType):
    pass


class InvalidUUID(InvalidStringType):
    pass


class InvalidNumberType(InvalidCastError):
    pass


class InvalidCurrency(InvalidNumberType):
    pass


class InvalidBooleanType(InvalidCastError):
    pass


class InvalidNoneType(InvalidCastError):
    pass


class InvalidObjectType(InvalidCastError):
    pass


class InvalidArrayType(InvalidCastError):
    pass


class InvalidDateType(InvalidCastError):
    pass


class InvalidDateTimeType(InvalidCastError):
    pass


class InvalidTimeType(InvalidCastError):
    pass


class InvalidGeoPointType(InvalidCastError):
    pass


class InvalidGeoJSONType(InvalidCastError):
    pass


# Constraints

class ConstraintError(JsonTableSchemaException):
    def __init__(self, msg=None):
        self.msg = msg or 'The value didn\'t validate against a constraint.'
        super(ConstraintError, self).__init__(msg)


class ConstraintNotSupported(JsonTableSchemaException):
    def __init__(self, msg=None):
        self.msg = msg or 'The field does not support the constraint.'
        super(ConstraintNotSupported, self).__init__(msg)


# Storage

class StorageError(JsonTableSchemaException):
    pass


# Deprecated

class ConversionError(JsonTableSchemaException):
    def __init__(self, msg=None):
        self.msg = msg or 'Error converting a row or field.'
        super(ConversionError, self).__init__(msg)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions
from .. import helpers
from . import base


# Module API

class BooleanType(base.JTSType):

    # Public

    name = 'boolean'
    null_values = helpers.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
    ]
    # ---
    python_type = bool
    true_values = helpers.TRUE_VALUES
    false_values = helpers.FALSE_VALUES

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            value = value.strip().lower()
        except AttributeError:
            pass

        if value in (self.true_values):
            return True
        elif value in (self.false_values):
            return False
        else:
            raise exceptions.InvalidBooleanType(
                '{0} is not a boolean value'.format(value))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


# Module API

def check_enum(name, value, enum):
    """Enum constraint.

    Check value is one from enumerable type.

    Args:
        name (str): field name
        value (any): field value
        enum (any[]): enum to check against

    Raises:
        TypeError: for non supported type
        ConstraintError: if check is failed

    """
    if value not in enum:
        raise exceptions.ConstraintError(
            "The value for field '{0}' must be in the enum array"
            .format(name))
    return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
from future.utils import raise_with_traceback
from .. import exceptions
from .. import helpers
from . import base


# Module API

class ObjectType(base.JTSType):

    # Public

    name = 'object'
    null_values = helpers.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minLength',
        'maxLength',
    ]
    # ---
    python_type = dict

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            json_value = json.loads(value)
            if isinstance(json_value, self.python_type):
                return json_value
            else:
                raise exceptions.InvalidObjectType()
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidObjectType(e))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import six
import decimal
import unicodedata
from future.utils import raise_with_traceback
from .. import exceptions
from .. import helpers
from .. import compat
from . import base


# Module API

class NumberType(base.JTSType):

    # Public

    name = 'number'
    null_values = helpers.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minimum',
        'maximum',
    ]
    # ---
    python_type = decimal.Decimal
    # We could change this initialization to something
    # more performant (http://bit.ly/1qgkK7B)
    currencies = u''.join(compat.chr(i) for i
                          in range(0xffff)
                          if unicodedata.category(compat.chr(i)) == 'Sc')

    def __preprocess_value(self, value):
        if type(value) is six.text_type:
            group_char = self.field.get('groupChar', ',')
            decimal_char = self.field.get('decimalChar', '.')
            percent_char = '%‰‱％﹪٪'
            value = value.replace(group_char, '').replace(decimal_char, '.')
            value = re.sub('['+percent_char+']', '', value)
        return value

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        if isinstance(value, (int, float)):
            return self.python_type(value)

        value = self.__preprocess_value(value)

        try:
            return self.python_type(value)
        except (ValueError, TypeError, decimal.InvalidOperation) as e:
            raise_with_traceback(exceptions.InvalidCastError(e))

        raise exceptions.InvalidCastError('Could not cast value')

    def cast_currency(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        if isinstance(value, (int, float)):
            return self.python_type(value)

        value = self.__preprocess_value(value)

        try:
            pattern = '[{0}]'.format(self.currencies)
            value = re.sub(pattern, '', value)
            return self.python_type(value)
        except (ValueError, TypeError, decimal.InvalidOperation):
            raise exceptions.InvalidCurrency(
                '{0} is not a valid currency'.format(value))

        raise exceptions.InvalidCastError('Could not cast value')



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from six import add_metaclass
from abc import ABCMeta, abstractmethod


# Module API

@add_metaclass(ABCMeta)
class Storage(object):
    """Tabular Storage interface.

    Concrete backends shouldn't inherit from this base class
    to simplify maintenance. Just use as a reference.

    Args:
        prefix (str): prefix for all buckets
        options (dict): concrete backend options

    """

    # Public

    @abstractmethod
    def __init__(self, prefix='', **options):
        pass

    @property
    def buckets(self):
        """str[]: list of buckets.

        This list should be sorted in order of foreign key dependency.
        (direct order for creation and reverse order for deletion)

        """
        pass

    @abstractmethod
    def create(self, bucket, descriptor, force=False):
        """Create one/few buckets.

        Args:
            bucket (str/list): bucket name or list of bucket names
            descriptor (dict/dict[]): schema descriptor or list of descriptors
            force (bool): delete and re-create already existent buckets

        Raises:
            StorageError: if table already exists.

        """
        pass

    def delete(self, bucket=None, ignore=False):
        """Delete one/few/all bucket(s).

        Args:
            bucket (str/list/None): bucket name or list of bucket
                names to delete. If None all buckets will be deleted.
            ignore (bool): don't raise an error on non-existent bucket
                deletion from storage

        Raises:
            StorageError: if table doesn't exist and ignore is False

        """
        pass

    def describe(self, bucket, descriptor=None):
        """Get/set bucket's schema descriptor.

        Args:
            table (str): bucket name
            schema (dict): schema descriptor to set

        Returns:
            dict: bucket's schema descriptor

        """
        pass

    def iter(self, bucket):
        """Yields bucket rows.

        This method should return typed values
        based on the schema of this buckets.

        Args:
            bucket (str): bucket name

        Yields:
            list: bucket row

        """
        pass

    def read(self, bucket):
        """Read rows from bucket.

        This method should return typed values
        based on the schema of this buckets.

        Args:
            bucket (str): bucket name

        Returns:
            list[]: bucket rows

        """
        pass

    def write(self, bucket, rows):
        """Write rows to bucket.

        This method should store values of unsupported
        types as strings internally (like csv does).

        Args:
            bucket (str): bucket name
            rows (list[]): rows to write

        """
        pass
